@@ -12,6 +12,8 @@ class TestFullPythonNetwork:
                  "c12-c12n-mg23-cf88",
                  "c12-c12p-na23-cf88",
                  "c12-ag-o16-nac2",
+                 "na23--ne23-toki",
+                 "ne23--na23-toki",
                  "n--p-wc12",
                  "he4-aag-c12-fy05"]
 
