@@ -5,12 +5,11 @@ particular, the binding energies)
 
 #__all__ = [binding_nuclide, binding_table, elements, partition_function]
 
-from .binding_nuclide import BindingNuclide
 from .binding_table import BindingTable
 from .elements import Element, PeriodicTable, UnidentifiedElement
-from .mass_nuclide import MassNuclide, MassTable
+from .mass_table import MassTable
 from .nucleus import Nucleus, UnsupportedNucleus
 from .partition_function import (PartitionFunction,
                                  PartitionFunctionCollection,
                                  PartitionFunctionTable)
-from .spin_nuclide import SpinNuclide, SpinTable
+from .spin_table import SpinTable
